#!/usr/bin/env python3
# -*- python-indent-offset: 4; -*-

from datetime import datetime, timedelta
from itertools import zip_longest

from dateutil import tz
from dateutil.parser import parse
from dateutil.rrule import DAILY, rrule
from taskw import TaskWarrior
from timew import Duration, TimeWarrior
from urwid import (
    Columns,
    ExitMainLoop,
    Filler,
    MainLoop,
    Pile,
    Text,
    raw_display,
    set_encoding,
)

from .config import Config
from .exceptions import (
    ArgumentException,
    IntervalIDException,
    IntervalModificationException,
    IntervalRangeException,
    IntervalTimeException,
    TaskNotFoundException,
    TimewIntervalNotFoundException,
)
from .model import Entries, Entry, Interval, Pomodoro
from .tdt import Complete, CompleteUnplanned, Empty, Interrupt, Void, VoidUnplanned


class Tasks:
    def __init__(self):
        self.config = Config().config
        self.taskw = TaskWarrior()
        #        self.taskw = TaskWarrior(config_filename="/home/tjaart/Code/pomw/testdata/taskrc")
        if self.config.getboolean("timew", "sync", fallback=True):
            self.timew = TimeWarrior()

        self.duration = self.config.getint("pomodoro", "length", fallback=25)

    def __parse_quantity(self, quantity):
        increment = False
        if quantity[0] == "+":
            increment = True
            quantity = int(quantity[1:])
        else:
            if quantity[0] == "-":
                increment = True
            quantity = int(quantity)

        return quantity, increment

    def __split_project(self, project):
        tags = []
        if "." in project:
            tags.extend([tag for tag in project.split(".")])
        tags.append(project)
        return tags

    def __find_interval_id(self, interval):
        entries = Entries()
        if not interval[0] == "@":
            raise IntervalIDException()

        interval_id = int(interval.strip("@"))
        if interval_id > len(entries.intervals()):
            raise IntervalRangeException(interval_id)

        return interval_id

    def __find_timew_id(self, interval):
        # TODO: We can limit the results for faster processing
        timew_list = self.timew.list()
        for list_entry in timew_list:
            if (
                parse(list_entry["start"]).astimezone(tz.gettz()).replace(tzinfo=None)
                == interval.start_time
                and parse(list_entry["end"]).astimezone(tz.gettz()).replace(tzinfo=None)
                == interval.end_time
            ):
                return list_entry["id"]

    def activity(self, action, date):
        print("activity")
        print(date)

    def default(self):
        print("default")

    def rm(self, action, interval):
        entries = Entries()
        interval = self.__find_interval_id(interval)

        interval_to_delete = entries.intervals()[interval - 1]
        for entry in entries:
            try:
                entry.intervals.remove(interval_to_delete)
            except ValueError:
                # we ignore value errors
                # TODO: this is not a very efficient algorithm... but OK for now
                pass

        if self.config.getboolean("timew", "sync", fallback=True):
            timew_id = self.__find_timew_id(interval_to_delete)

            if not timew_id:
                raise TimewIntervalNotFoundException(interval_to_delete)

            self.timew.delete(timew_id)

        entries.save()

    def interrupt(self, action, time, type, task, quantity):
        # TODO: The current implementation only allows for interruptions in the current, unfinished pomodoro
        #       To retroactively add interruptions will require some more work.

        task_id, cur_task = self.taskw.get_task(id=task)
        quantity, increment = self.__parse_quantity(quantity)

        if not task_id:
            raise TaskNotFoundException(task)

        entries = Entries()
        entry = (
            entries.find(date=time.date(), uuid=cur_task["uuid"])[0]
            if entries.find(date=time.date(), uuid=cur_task["uuid"])
            else None
        )
        if entry:
            # If the last pomodoro already has a completed time, create a new pomodoro for the interruption
            if len(entry.pomodoros) == 0 or entry.pomodoros[-1].end_time:
                # TODO: Simplify internal/external if block
                if type == "internal":
                    entry.interval = entry.intervals + [
                        Pomodoro(internal_interrupt=quantity)
                    ]
                else:
                    entry.intervals = entry.intervals + [
                        Pomodoro(external_interrupt=quantity)
                    ]
            else:
                if type == "internal":
                    if increment:
                        entry.pomodoros[-1].internal_interrupt += quantity
                    else:
                        entry.pomodoros[-1].internal_interrupt = quantity
                else:
                    if increment:
                        entry.pomodoros[-1].external_interrupt += quantity
                    else:
                        entry.pomodoros[-1].external_interrupt = quantity

        else:
            # TODO: Simplify internal/external if block
            if type == "internal":
                entry = Entry(
                    uuid=cur_task["uuid"],
                    date=time.date(),
                    pomodoros=[Pomodoro(internal_interrupt=quantity)],
                )
            else:
                entry = Entry(
                    uuid=cur_task["uuid"],
                    date=time.date(),
                    pomodoros=[Pomodoro(external_interrupt=quantity)],
                )
        entries.entries = entries.entries + [entry]
        entries.save()

    def done(self, action, task, end_time):
        task_id, cur_task = self.taskw.get_task(id=task)

        if not task_id:
            raise TaskNotFoundException(task)

        entries = Entries()
        entry = (
            entries.find(date=end_time.date(), uuid=cur_task["uuid"])[0]
            if entries.find(date=end_time.date(), uuid=cur_task["uuid"])
            else None
        )

        if not entry:
            entries.entries += [
                Entry(
                    uuid=cur_task["uuid"],
                    date=end_time.date(),
                    intervals=[
                        Pomodoro(
                            start_time=(end_time - timedelta(minutes=self.duration)),
                            end_time=end_time,
                        )
                    ],
                )
            ]
        else:
            last_pomodoro = None
            if entry and entry.pomodoros:
                last_pomodoro = entry.pomodoros[-1]

            if (
                not last_pomodoro
                or last_pomodoro.void
                or last_pomodoro.end_time is not None
            ):
                entry.intervals.append(
                    Pomodoro(
                        start_time=(end_time - timedelta(minutes=self.duration)),
                        end_time=end_time,
                    )
                )
            else:
                last_pomodoro.end_time = end_time

        tags = [cur_task.get("description", "")] + self.__split_project(
            cur_task.get("project", "")
        )

        if self.config.getboolean("timew", "sync", fallback=True):
            self.timew.track(
                end_time - timedelta(minutes=self.duration), end_time, tags=tags
            )

        entries.save()

    def plan(self, action, task, quantity, date):
        # get the corresponding Taskwarrior task
        task_id, cur_task = self.taskw.get_task(id=task)

        if not task_id:
            raise TaskNotFoundException(task)

        quantity, increment = self.__parse_quantity(quantity)

        entries = Entries()
        entry = (
            entries.find(date=date, uuid=cur_task["uuid"])[0]
            if entries.find(date=date, uuid=cur_task["uuid"])
            else None
        )

        if not entry:
            entry = Entry(uuid=cur_task["uuid"], date=date)
            entries.entries += [entry]

        if increment:
            entry.planned += quantity
        else:
            entry.planned = quantity

        entries.save()

    def start(self, action):
        print(action)
        # data = json.loads(Popen([cmd, 'export'], stdout=PIPE).stdout.read())

    def void(self, action, task, date):
        task_id, cur_task = self.taskw.get_task(id=task)

        if not task_id:
            raise TaskNotFoundException(task)

        entries = Entries()
        entry = (
            entries.find(date=date, uuid=cur_task["uuid"])[0]
            if entries.find(date=date, uuid=cur_task["uuid"])
            else None
        )

        if not entry:
            entry = Entry(uuid=cur_task["uuid"], date=date)
            entries.entries += [entry]

        # If last pomodoro is complete, add new voided pomodoro, else void the existing pomodoro
        if not entry.pomodoros:
            entry.intervals += [Pomodoro(void=True)]
        if not entry.pomodoros[-1].end_time:
            entry.pomodoros[-1].void = True
        else:
            entry.intervals += [Pomodoro(void=True)]

        entries.save()

    def add(self, action, task, start_time, end_time, duration):
        task_id, cur_task = self.taskw.get_task(id=task)
        if not task_id:
            raise TaskNotFoundException(task)

        tags = [cur_task.get("description", "")] + self.__split_project(
            cur_task.get("project", "")
        )

        entries = Entries()
        entry = entries.find(uuid=cur_task["uuid"], date=end_time.date())

        if entry:
            entry = entry[0]
        else:
            entry = Entry(uuid=cur_task["uuid"], date=end_time.date())
            entries.entries += [entry]

        if end_time and duration:
            start_time = end_time - timedelta(minutes=duration)
        elif start_time and duration:
            end_time = start_time + timedelta(minutes=duration)
        elif start_time and end_time:
            pass
        else:
            raise ArgumentException(
                "Two of END_TIME, START_TIME and DURATION must be provided!"
            )

        entry.intervals += [Interval(start_time=start_time, end_time=end_time)]
        if self.config.getboolean("timew", "sync", fallback=True):
            self.timew.track(start_time, end_time=end_time, tags=tags)

        entries.save()

    def ls(self, action, date):
        entries = Entries()

        # underlined_headers = ''.join(tuple(x + y for x, y in zip_longest(
        #     'Type   ID    Start Time      End Time', '\u0332', fillvalue='\u0332')))
        id_h = "".join(tuple(x + y for x, y in zip_longest("ID", "̲", fillvalue="̲")))
        project_h = "".join(
            tuple(x + y for x, y in zip_longest("Project", "̲", fillvalue="̲"))
        )
        task_h = "".join(
            tuple(x + y for x, y in zip_longest("Task", "̲", fillvalue="̲"))
        )
        type_h = "".join(
            tuple(x + y for x, y in zip_longest("Type", "̲", fillvalue="̲"))
        )
        date_h = "".join(
            tuple(x + y for x, y in zip_longest("Date", "̲", fillvalue="̲"))
        )
        start_h = "".join(
            tuple(x + y for x, y in zip_longest("Start", "̲", fillvalue="̲"))
        )
        end_h = "".join(tuple(x + y for x, y in zip_longest("End", "̲", fillvalue="̲")))

        if len(entries.intervals()) > 0:
            print(
                "{:6} {:11} {:15} {:14} {:12} {:29} {}".format(
                    id_h, type_h, date_h, start_h, end_h, project_h, task_h
                )
            )
            interval_id = 1
            for interval in entries.intervals():
                task_id, cur_task = self.taskw.get_task(id=interval.task)

                interval_type = "Pom" if isinstance(interval, Pomodoro) else "NonPom"
                print(
                    "\033[91m@{:<3}\033[00m {:7} {!s:11} {!s:9} {!s:9} {!s:22} {!s}".format(
                        interval_id,
                        interval_type,
                        interval.start_time.date(),
                        interval.start_time.time(),
                        interval.end_time.time(),
                        cur_task.get("project", None),
                        cur_task.get("description", None),
                    )
                )
                interval_id += 1
        else:
            print("No entries to display")

    def mod(self, action, interval, start_time, end_time):
        entries = Entries()

        interval = self.__find_interval_id(interval)
        interval_to_modify = entries.intervals()[interval - 1]

        if isinstance(interval_to_modify, Pomodoro):
            raise IntervalModificationException

        original_duration = interval_to_modify.end_time - interval_to_modify.start_time
        timew_id = self.__find_timew_id(interval_to_modify)

        if not start_time and not end_time:
            print("Nothing modified. Set a new start_time or end_time")
        if start_time:
            interval_to_modify.start_time = start_time
        if end_time:
            interval_to_modify.end_time = end_time

        new_duration = interval_to_modify.end_time - interval_to_modify.start_time

        if interval_to_modify.start_time > interval_to_modify.end_time:
            raise IntervalTimeException

        # Update timewarrior
        # If the new interval is shorter, first shorten it, then move start,
        # if it is longer, we first move then lengthen
        if self.config.getboolean("timew", "sync", fallback=True):
            if not timew_id:
                raise TimewIntervalNotFoundException(interval_to_modify)
            if original_duration > new_duration:
                self.timew.shorten(timew_id, Duration(original_duration - new_duration))
                if start_time:
                    self.timew.move(timew_id, start_time)
            else:
                if start_time:
                    self.timew.move(timew_id, start_time)
                self.timew.lengthen(
                    timew_id, Duration(new_duration - original_duration)
                )

        entries.save()

    def mv(self, action, interval, start_time):
        entries = Entries()

        interval_to_mv = entries.intervals()[self.__find_interval_id(interval) - 1]
        timew_id = self.__find_timew_id(interval_to_mv)
        duration = interval_to_mv.end_time - interval_to_mv.start_time
        interval_to_mv.start_time = start_time
        interval_to_mv.end_time = start_time + duration

        if self.config.getboolean("timew", "sync", fallback=True):
            if not timew_id:
                raise TimewIntervalNotFoundException(interval_to_mv)
            self.timew.move(timew_id, start_time)

        entries.save()

    def tdt(self, action, date):
        def unhandled_input(key):
            if key in ("Q", "q", "esc"):
                raise ExitMainLoop()

        entries = Entries()
        data = entries.find(date=date)
        render_pomodoros = []
        render_nonpom = []
        nonpom_to_render = False
        grand_total_pom = timedelta()
        grand_total_nonpom = timedelta()

        for entry in data:
            task_id, task = self.taskw.get_task(uuid=entry.uuid)
            if not task_id:
                task_id = "-"

            project = task["project"]
            description = task["description"]
            planned = entry.planned
            pomodoros = entry.pomodoros
            nonpomodoros = entry.nonpomodoros

            rendered_pomodoros = []
            if planned or pomodoros:
                # TODO: Simplify the code here!
                if planned >= len(pomodoros):
                    for i in range(0, len(pomodoros)):
                        if pomodoros[i].void:
                            rendered_pomodoros.append((3, Void()))
                        else:
                            rendered_pomodoros.append((3, Complete()))
                            grand_total_pom += timedelta(minutes=self.duration)
                    for i in range(0, planned - len(pomodoros)):
                        rendered_pomodoros.append((3, Empty()))
                else:
                    for i in range(0, planned):
                        if pomodoros[i].void:
                            rendered_pomodoros.append((3, Void()))
                        else:
                            rendered_pomodoros.append((3, Complete()))
                            grand_total_pom += timedelta(minutes=self.duration)
                    for i in range(planned, len(pomodoros)):
                        if pomodoros[i].void:
                            rendered_pomodoros.append((3, VoidUnplanned()))
                        else:
                            rendered_pomodoros.append((3, CompleteUnplanned()))
                            grand_total_pom += timedelta(minutes=self.duration)

                row = Columns(
                    [
                        Filler(Text("{:>3}  {}".format(task_id, description))),
                        Columns(rendered_pomodoros),
                    ]
                )
                render_pomodoros.append((3, row))

            if nonpomodoros:
                nonpom_to_render = True
                total_time = timedelta()

                for i in range(0, len(nonpomodoros)):
                    total_time += nonpomodoros[i].end_time - nonpomodoros[i].start_time
                    grand_total_nonpom += total_time

                row = Columns(
                    [
                        Filler(Text("{:>3}  {}".format(task_id, description))),
                        Filler(Text("{}".format(total_time))),
                    ]
                )
                render_nonpom.append((2, row))

        pom_header = (
            "pack",
            Columns(
                [
                    Text("To Do Today\n-----------"),
                    Text("{}\n----------".format(date), align="right"),
                ]
            ),
        )
        if grand_total_pom:
            pom_footer = (
                "pack",
                Columns([Text(""), Text("-------\n{}".format(grand_total_pom))]),
            )
        else:
            pom_footer = ("pack", Columns([Text("")]))

        nonpom_header = ("pack", Columns([Text("")]))
        nonpom_footer = ("pack", Columns([Text("")]))
        if nonpom_to_render:
            nonpom_header = (
                "pack",
                Columns([Text("\nNon Pomodoro Time\n-----------------")]),
            )
            nonpom_footer = (
                "pack",
                Columns([Text(""), Text("-------\n{}".format(grand_total_nonpom))]),
            )

        # footer = ('pack', Text('Footer'))

        widget = Pile(
            [pom_header]
            + render_pomodoros
            + [pom_footer]
            + [nonpom_header]
            + render_nonpom
            + [nonpom_footer]
        )

        loop = MainLoop(
            widget,
            screen=raw_display.Screen(),
            unhandled_input=unhandled_input,
            pop_ups=True,
        )
        loop.run()



# -*- python-indent-offset: 4; -*-


class IntervalIDException(Exception):
    def __init__(self):
        super().__init__("Interval ids start with a '@' character.")


class TimewIntervalNotFoundException(Exception):
    def __init__(self, interval):
        super().__init__(
            f"Timewarrior interval: {interval.start_time} - {interval.end_time} does not exist."
        )


class IntervalTimeException(Exception):
    def __init__(self):
        super().__init__("Interval start_time past end_time.")


class IntervalModificationException(Exception):
    def __init__(self):
        super().__init__("Pomodoro intervals cannot be modified.")


class IntervalRangeException(Exception):
    def __init__(self, interval_id):
        super().__init__("Interval @%d does not exist." % (interval_id))


class TaskNotFoundException(Exception):
    def __init__(self, task_id):
        super().__init__("Task id %d could not be found." % (task_id))


class EntryFoundException(Exception):
    def __init__(self, uuid):
        super().__init__("Entry with uuid %s could not be found." % (uuid))


class ArgumentException(Exception):
    def __init__(self, message):
        super().__init__(message)



# -*- python-indent-offset: 4; -*-

from urwid import Filler, Text, WidgetWrap


class Empty(WidgetWrap):
    def __init__(self):
        box = "┌─┐\n│ │\n└─┘"
        super().__init__(Filler(Text(box)))


class Complete(WidgetWrap):
    def __init__(self):
        box = "┌─┐\n│x│\n└─┘"
        super().__init__(Filler(Text(box)))


class CompleteUnplanned(WidgetWrap):
    def __init__(self):
        box = "   \n x \n   "
        super().__init__(Filler(Text(box)))


class Interrupt(WidgetWrap):
    def __init__(self):
        box = "┌─┐\n│'│\n└─┘"
        super().__init__(Filler(Text(box)))


class Void(WidgetWrap):
    def __init__(self):
        box = "┌─┐\n│/│\n└─┘"
        super().__init__(Filler(Text(box)))


class VoidUnplanned(WidgetWrap):
    def __init__(self):
        box = "   \n / \n   "
        super().__init__(Filler(Text(box)))



# -*- python-indent-offset: 4; -*-

import calendar
from datetime import datetime
from os import listdir
from os.path import abspath, expanduser, getsize, isfile, join, splitext

import unijson
from dateutil.parser import parse
from dateutil.rrule import DAILY, MONTHLY, rrule

from .entry import Entry


class Entries(object):
    """A collection of Pomwarrior entries.

    This class implements the Active Record pattern:
    Entries contains a month of sorted list of Entry Objects. (Entry is sorted by its Date and UUID)

    Args:
        path (str): The path to the data files directory
        entries (list of :Entry:): A list of entries (use this when not loading data from files)
        load_from_file (bool): Should the data be loaded from the filesystem
        filter (date): Load data after the given date
    """

    def __init__(self, path="~/.config/pomw/data", entries=[], load=True):
        self.__path = path
        self.__entries = []
        if load:
            abs_path = abspath(expanduser(path))
            for file_name in sorted(listdir(abs_path)):
                # ignore non-json files
                if splitext(file_name)[1] == ".json":
                    loaded = Entries.load(join(abs_path, file_name))
                    self.entries += loaded
        else:
            self.entries = entries

    @property
    def entries(self):
        return self.__entries

    @entries.setter
    def entries(self, entries):
        self.__entries = []
        if not isinstance(entries, list):
            raise TypeError("'entries' should be a %s" % (list.__name__))

        for entry in entries:
            if isinstance(entry, dict):
                self.entries.append(
                    Entry(
                        uuid=entry.get("uuid"),
                        date=entry.get("date"),
                        planned=entry.get("planned", 0),
                        intervals=entry.get("intervals", []),
                    )
                )
            elif isinstance(entry, Entry):
                self.entries.append(entry)
            else:
                raise TypeError(
                    '"entries" elements should be of type %s' % (Entry.__name__)
                )
        self.entries.sort()

    def __len__(self):
        return len(self.entries)

    def __getitem__(self, key):
        return self.entries[key]

    def find(self, uuid=None, date=None):
        """Find the matching Entry for the given uuid and date

        Arguments:
            uuid (str) or list of str: The uuid(s) to match
            date (datetime.date) or list of `datetime.date`: The date(s) to match

        Returns:
            A list of Entries matching both criteria (AND) is returned
        """
        uuid_criteria = None
        date_criteria = None
        if uuid:
            uuid_criteria = uuid if isinstance(uuid, list) else [uuid]

        if date:
            date_criteria = date if isinstance(date, list) else [date]

        return_list = []

        for entry in self.entries:
            if (not uuid_criteria or entry.uuid in uuid_criteria) and (
                not date_criteria or entry.date in date_criteria
            ):
                return_list += [entry]

        return return_list

    def insert(self, entry):
        self.entries.append(entry)

    @staticmethod
    def load(file_name):
        """Load Pomodoro data from a file

        Arguments: file_name (str): The file to load the data from

        Returns:
            List of Entry objects
        """
        try:
            fp = open(file_name, "r")

            # edge case when file has been created, but is empty
            if getsize(file_name) == 0:
                return []

            return unijson.load(fp)
        except FileNotFoundError:
            return []

    def save(self, path=None):
        """Save Pomodoro data to a file

        Arguments: path (str): The file to save the data to

        Returns:
            List of Entry objects
        """
        if not path:
            path = self.__path

        first_month = datetime(self.entries[0].date.year, self.entries[0].date.month, 1)
        last_month = datetime(
            self.entries[-1].date.year,
            self.entries[-1].date.month,
            calendar.monthrange(
                self.entries[-1].date.year, self.entries[-1].date.month
            )[1],
            23,
            59,
            59,
        )

        for dt in rrule(MONTHLY, dtstart=first_month, until=last_month):
            days = [
                dt2.date()
                for dt2 in list(
                    rrule(
                        DAILY,
                        dtstart=dt,
                        count=calendar.monthrange(dt.year, dt.month)[1],
                    )
                )
            ]

            entries = self.find(date=days)

            file_name = join(
                abspath(expanduser(path)), "%s.json" % (dt.strftime("%Y-%m"))
            )
            with open(file_name, "w") as file:
                unijson.dump(entries, file)

    def intervals(self):
        intervals = []
        for entry in self.entries:
            intervals += entry.intervals
        intervals.sort(reverse=True)
        return intervals

    def __repr__(self):
        return_text = "[\n"
        for entry in self.entries:
            return_text += "  " + str(entry) + "\n"
        return_text += "]"
        return return_text



#!/usr/bin/env python3

import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, ArgumentTypeError
from datetime import date, datetime, timedelta

# from argcomplete import autocomplete
from dateutil.parser import parse

from pomw.exceptions import (
    ArgumentException,
    EntryFoundException,
    IntervalIDException,
    IntervalModificationException,
    IntervalRangeException,
    IntervalTimeException,
    TaskNotFoundException,
    TimewIntervalNotFoundException,
)
from pomw.tasks import Tasks


class Application:
    def valid_date(self, s):
        try:
            return parse(s).date()
        except ValueError:
            msg = "Not a valid date: '{0}'.".format(s)
            raise ArgumentTypeError(msg)

    def valid_datetime(self, s):
        try:
            return parse(s)
        except ValueError:
            msg = "Not a valid date: '{0}'.".format(s)
            raise argparse.ArgumentTypeError(msg)

    def create_parser(self, program_bin):
        parser = ArgumentParser(
            prog=program_bin, formatter_class=ArgumentDefaultsHelpFormatter
        )
        # autocomplete(parser)
        subparsers = parser.add_subparsers(dest="action")

        plan_parser = subparsers.add_parser(
            "plan",
            help="Plan a pomodoro",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        plan_parser.add_argument("task", help="Task id", type=int)
        plan_parser.add_argument(
            "-q",
            "--quantity",
            default="+1",
            help="The number of pomodoros (default: +1)",
            type=str,
        )
        plan_parser.add_argument(
            "-d", "--date", default=date.today(), help="The date", type=self.valid_date
        )

        complete_parser = subparsers.add_parser(
            "done",
            help="Complete a pomodoro",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        complete_parser.add_argument("task", help="Task id", type=int)
        complete_parser.add_argument(
            "-e",
            "--end_time",
            help="The end time for the pomodoro",
            default=(datetime.now()).strftime("%Y-%m-%dT%H:%M:%S"),
            type=self.valid_datetime,
        )

        void_parser = subparsers.add_parser(
            "void",
            help="Void a pomodoro",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        void_parser.add_argument("task", help="Task id", type=int)
        void_parser.add_argument(
            "-d", "--date", default=date.today(), help="The date", type=self.valid_date
        )

        #       pom interrupt [internal|external] Mark an interruption for the current pomodoro
        interrupt_parser = subparsers.add_parser(
            "interrupt", help="Interrupt the current pomodoro"
        )
        interrupt_parser.add_argument(
            "type",
            default="internal",
            help="The interruption type",
            choices=["internal", "external"],
        )
        interrupt_parser.add_argument("task", help="Task id", type=int)
        interrupt_parser.add_argument(
            "-t",
            "--time",
            default=(datetime.now()).strftime("%Y-%m-%dT%H:%M:%S"),
            help="The interruption time",
            type=self.valid_datetime,
        )
        interrupt_parser.add_argument(
            "-q",
            "--quantity",
            default="+1",
            help="The number of interruptions (default: +1)",
            type=str,
        )

        list_parser = subparsers.add_parser(
            "ls", help="List intervals", formatter_class=ArgumentDefaultsHelpFormatter
        )
        list_parser.add_argument(
            "-d",
            "--date",
            help="The date for entries",
            default=(datetime.now()).strftime("%Y-%m-%d"),
            type=self.valid_date,
        )

        nonpom_parser = subparsers.add_parser(
            "add",
            help="Add a time interval",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        nonpom_parser.add_argument("task", help="Task id", type=int)
        nonpom_parser.add_argument(
            "-s",
            "--start_time",
            help="The start time for the time entry. If both END_TIME and DURATION \
    is given, START_TIME gets ignored",
            type=self.valid_datetime,
        )
        nonpom_parser.add_argument(
            "-e",
            "--end_time",
            help="The end time for the time entry",
            default=(datetime.now()).strftime("%Y-%m-%dT%H:%M:%S"),
            type=self.valid_datetime,
        )
        nonpom_parser.add_argument(
            "-d",
            "--duration",
            help="The duration for the time entry in minutes.",
            default=30,
            type=int,
        )

        modify_parser = subparsers.add_parser(
            "mod",
            help="Modify an interval",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        modify_parser.add_argument("interval", help="Interval id", type=str)
        modify_parser.add_argument(
            "-s",
            "--start_time",
            help="The new start time for the interval",
            type=self.valid_datetime,
        )
        modify_parser.add_argument(
            "-e",
            "--end_time",
            help="The new end time for the interval",
            type=self.valid_datetime,
        )

        move_parser = subparsers.add_parser(
            "mv", help="Move an interval", formatter_class=ArgumentDefaultsHelpFormatter
        )
        move_parser.add_argument("interval", help="Interval id", type=str)
        move_parser.add_argument(
            "-s",
            "--start_time",
            help="The new start time for the interval",
            type=self.valid_datetime,
        )

        remove_parser = subparsers.add_parser(
            "rm",
            help="Remove an interval",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        remove_parser.add_argument("interval", help="Interval id", type=str)

        # start_parser = subparsers.add_parser('start',
        #                                      help='Start a new pomodoro')
        # start_parser.add_argument('task',
        #                           help='Task id',
        #                           type=int)

        tdt_parser = subparsers.add_parser(
            "tdt",
            help="Print To Do Today sheet",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        tdt_parser.add_argument(
            "-d",
            "--date",
            help="Date to print",
            default=datetime.now().strftime("%Y-%m-%d"),
            type=self.valid_date,
        )

        return parser

    def run(self):
        try:
            tasks = Tasks()
            parser = self.create_parser(sys.argv[0])
            args = parser.parse_args(sys.argv[1:])
            action = args.action
            kwargs = vars(args)
            # del kwargs['action']
            if not action:
                parser.print_help(file=None)
            else:
                method_to_call = getattr(tasks, action)
                result = method_to_call(**kwargs)
        except (
            IntervalIDException,
            IntervalModificationException,
            IntervalRangeException,
            IntervalTimeException,
            TaskNotFoundException,
            TimewIntervalNotFoundException,
            EntryFoundException,
            ArgumentException,
        ) as e:
            print("Error: %s" % (e), file=sys.stderr)
            exit(1)



# -*- python-indent-offset: 4; -*-

from abc import ABC
from datetime import datetime


class Interval(ABC):
    def __init__(self, start_time=None, end_time=None):
        self.start_time = start_time
        self.end_time = end_time

    @staticmethod
    def __json_decode__(d):
        return Interval(
            start_time=d.get("_Interval__start_time", None),
            end_time=d.get("_Interval__end_time", None),
        )

    @property
    def start_time(self):
        return self.__start_time

    @start_time.setter
    def start_time(self, start_time):
        if isinstance(start_time, datetime):
            self.__start_time = start_time
        elif not start_time:
            self.__start_time = None
        else:
            raise TypeError("'start_time' should be of type %s" % (datetime.__name__))

    @property
    def end_time(self):
        return self.__end_time

    @end_time.setter
    def end_time(self, end_time):
        if isinstance(end_time, datetime):
            self.__end_time = end_time
        elif not end_time:
            self.__end_time = None
        else:
            raise TypeError("'end_time' should be of type %s" % (datetime.__name__))

    def __lt__(self, other):
        self_start = datetime.min
        self_end = datetime.min
        other_start = datetime.min
        other_end = datetime.min
        if self.start_time:
            self_start = self.start_time
        if self.end_time:
            self_end = self.end_time
        if other.start_time:
            other_start = other.start_time
        if other.end_time:
            other_end = other.end_time

        if self_start == other_start:
            return self_end < other_end
        return self_start < other_start

    def __repr__(self):
        return "{start_time: %s, end_time: %s}" % (self.start_time, self.end_time)



# -*- python-indent-offset: 4; -*-

from configparser import ConfigParser
from os.path import expanduser


class Config:
    def __init__(self, config_filename="%s/.config/pomw/pomwrc" % expanduser("~")):
        self.__config = ConfigParser()
        self.__config.read(config_filename)

    @property
    def config(self):
        return self.__config



# -*- python-indent-offset: 4; -*-

import datetime

import unijson

from .interval import Interval
from .pomodoro import Pomodoro


class Entry:
    """A Pomwarrior entry"""

    def __init__(
        self, uuid=None, date=datetime.datetime.now().date(), planned=0, intervals=[]
    ):
        self.uuid = uuid
        self.date = date
        self.planned = planned
        self.intervals = intervals

    @staticmethod
    def __json_decode__(d):
        intervals = []
        for c in d.get("_Entry__intervals", []):
            intervals.append(c)

        entry = Entry(
            uuid=d.get("_Entry__uuid", None),
            date=d.get("_Entry__date", None),
            planned=d.get("_Entry__planned", 0),
            intervals=intervals,
        )

        return entry

    @property
    def uuid(self):
        return self.__uuid

    @uuid.setter
    def uuid(self, uuid):
        self.__uuid = uuid

    @property
    def date(self):
        return self.__date

    @date.setter
    def date(self, date):
        if not isinstance(date, datetime.date):
            raise TypeError("'date' should be of type %s" % (datetime.date.__name__))
        self.__date = date

    @property
    def planned(self):
        return self.__planned

    @planned.setter
    def planned(self, planned):
        self.__planned = planned

    @property
    def intervals(self):
        return self.__intervals

    @property
    def pomodoros(self):
        return list(
            filter(lambda interval: isinstance(interval, Pomodoro), self.__intervals)
        )

    @property
    def nonpomodoros(self):
        return list(
            filter(
                lambda interval: isinstance(interval, Interval)
                and not isinstance(interval, Pomodoro),
                self.__intervals,
            )
        )

    @intervals.setter
    def intervals(self, intervals):
        self.__intervals = []
        for interval in intervals:
            if isinstance(interval, dict):
                self.__entries.append(
                    Interval(
                        task=self.uuid,
                        start=interval.get("start", None),
                        end=interval.get("end", None),
                    )
                )
            elif isinstance(interval, Interval):
                interval.task = self.uuid
                self.__intervals.append(interval)
            else:
                raise TypeError('Unable to parse "interval"')

    def __repr__(self):
        return "{date: %s, uuid: %s, planned: %s, intervals: %s }" % (
            self.date,
            self.uuid,
            self.planned,
            self.intervals,
        )

    def __eq__(self, other):
        return self.date == other.date and self.uuid == other.uuid

    def __lt__(self, other):
        if self.date == other.date:
            return self.uuid < other.uuid
        return self.date < other.date



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['pomw', 'pomw.console', 'pomw.model']

package_data = \
{'': ['*']}

install_requires = \
['python-dateutil>=2.7,<3.0',
 'semver>=2.8,<3.0',
 'taskw>=1.2,<2.0',
 'timew>=0.0.22,<0.0.23',
 'unijson>=1.0,<2.0',
 'urwid>=2.0,<3.0']

entry_points = \
{'console_scripts': ['pomw = pomw.console:main']}

setup_kwargs = {
    'name': 'pomw',
    'version': '0.1.0',
    'description': 'The Pomodoro Technique using TaskWarrior and TimeWarrior',
    'long_description': '# PomodoroWarrior #\n\nA command line tool that integrates with TaskWarrior and TimeWarrior to implement the Pomodoro Technique.\n\n## Configuration ##\n\nThe configuration file is located at `$HOME/.config/pomw/pomwrc`\n\n### Timewarrior integration ###\n\nPomodoroWarrior aims to play nicely with TimeWarrior. By default all intervals created by PomodoroWarrior will also be updated in TimeWarrior. To change the default value, add the following section to your configuration file:\n \n```ini\n[timew]\nsync = false\n```\n\n### Pomodoro length ###\n\nPomodoro lengths are also configurable. Default length is 25 minutes.\n\n```ini\n[pomw]\nlength = 30\n```\n\n## Command Line Interface ##\n\nPomodoroWarrior uses a command line interface similar to TaskWarrior and TimeWarrior.\n\nCommands can refer to both TaskWarrior task ids, and PomodoroWarrior interval ids. To make the distinction clear, we force interval ids to be prepended by a `@` sign.\n\nThe commands can be grouped into 2 categories. Those that deal with Pomodoros, and those that deal with generic time intervals.\n\n### Workflow example ###\n\nFirst thing in the morning, we plan our Pomodoros for the day. We view our TaskWarrior task list.\n\n```bash\n$ task next\n```\n\n```\nID Age  P Project      Tag  Due Description                      Urg \n1 6w   M pomw                2d Create documentation for pomw     10\n2 6w   M python              1w Find the holy grail                8\n```\n\nWe decide to plan 5 pomodoros for writing this documentation\n\n```bash\n$ pomw plan -q 4 1 # Plan 4 pomodoros for task 1\n```\n\nWe check our To Do Today sheet\n\n```bash\n$ pomw tdt # Print the To Do Today sheet\n```\n\n```\nTo Do Today                                                        2019-03-21\n-----------                                                        ----------\n                                     ┌─┐┌─┐┌─┐┌─┐\n1  Create documentation for pomw     │ ││ ││ ││ │\n                                     └─┘└─┘└─┘└─┘\n```\n\nOops, we wanted 5 pomodoros, but we only planned 4. We now have 2 options, we can either override the number of planned pomodoros\n\n```bash\n$ pomw plan -q 5 1 # Plan 5 pomodoros for task 1\n```\n\nor append more pomodoros to the existing number of planned pomodoros\n\n```bash\n$ pomw plan -q +1 1 # Increment the planned pomodoros for task 1 by 1\n```\n\nWe check our tdt sheet again, and all is good!\n\n```\nTo Do Today                                                        2019-03-21\n-----------                                                        ----------\n                                     ┌─┐┌─┐┌─┐┌─┐┌─┐\n1  Create documentation for pomw     │ ││ ││ ││ ││ │\n                                     └─┘└─┘└─┘└─┘└─┘\n```\n\nSince we cannot find the holy grail using pomodoros, we will add the time we spend on this task using (non-pomodoro) intervals. \n\nOur To Do Today sheet looks good, and we start working on our first pomodoro.\n\n<!-- After working for several minutes, we remember that we need to buy SPAM. We add this to our shopping list, and mark an internal interruption -->\n\n<!-- ``` -->\n<!-- $ pomw interrupt internal 1 -->\n<!-- ``` -->\n\n<!-- Now our To Do Today sheet looks like this -->\n\n<!-- ``` -->\n<!-- To Do Today                                                        2019-03-21 -->\n<!-- -----------                                                        ---------- -->\n<!--                                      ┌─┐┌─┐┌─┐┌─┐┌─┐ -->\n<!-- 1  Create documentation for pomw     │ │ │ │ │ │ │ │ │ │ -->\n<!--                                      └─┘└─┘└─┘└─┘└─┘ -->\n<!-- ``` -->\n\nAfter working for several minutes, we start dreaming about lady Gwenevere, and before we realize it 10 minutes have passed. Because we believe in strictly following the Pomodoro Technique, we void the pomodoro.\n\n```bash\n$ pomw void 1 # Void the current pomodoro for task 1\n```\n\nOur TDT sheet reflects the voided Pomodoro\n\n```\nTo Do Today                                                        2019-03-21\n-----------                                                        ----------\n                                     ┌─┐┌─┐┌─┐┌─┐┌─┐\n1  Create documentation for pomw     │/││ ││ ││ ││ │\n                                     └─┘└─┘└─┘└─┘└─┘\n```\n\nOn our second attempt, we complete a pomodoro. Once it rings, we mark the pomodoro as done.\n\n```bash\n$ pomw done 1 # Complete a pomodoro for task 1\n```\n\n```\nTo Do Today                                                        2019-03-21\n-----------                                                        ----------\n                                     ┌─┐┌─┐┌─┐┌─┐┌─┐\n1  Create documentation for pomw     │/││x││ ││ ││ │\n                                     └─┘└─┘└─┘└─┘└─┘\n                                      --------\n                                      00:25:00\n```\n\nWe search for the grail. After an hour we give up to take a nap. We log our time as an interval\n\n```bash\n$ pomw add -d 60 2 # Add a 60 minute time interval for task 2\n```\n\n```\nTo Do Today                                                        2019-03-21\n-----------                                                        ----------\n                                     ┌─┐┌─┐┌─┐┌─┐┌─┐\n1  Create documentation for pomw     │/││x││ ││ ││ │\n                                     └─┘└─┘└─┘└─┘└─┘\n                                      -------\n                                      0:25:00\n                                      \nNon Pomodoro Time                     \n-----------------\n2  Find the holy grail                1:00:00\n                                      -------\n                                      1:00:00\n```\n\nKing Arthur will only pay for 50 minutes of our time, so we modify the time interval\n\n```bash\n$ pomw ls  # Lists the time intervals contained in the database\n```\n\n```\nID   Type    Date        Start     End       Project                Task\n@1   NonPom  2019-03-21  13:43:32  14:43:32  python                 Find the holy grail\n@2   Pom     2019-03-21  13:15:17  13:40:17  pomw                   Create documentation for pomw\n```\n\nFrom the listing, we see that we need to modify interval `@1`. We set the new end time to 14:33:32\n\n```bash\n$ pomw mod -e 14:33:32 @1 # Modify interval @1, set new end time as 14:33:32\n```\n\n```\nID   Type    Date        Start     End       Project                Task\n@1   NonPom  2019-03-21  13:43:32  14:33:32  python                 Find the holy grail\n@2   Pom     2019-03-21  13:15:17  13:40:17  pomw                   Create documentation for pomw\n```\n\nKing Arthur will not pay for our documentation creation efforts. We delete the pomodoro at interval `@2`.\n\n```bash\n$ pomw rm @2 # Delete interval @2 \n```\n\n```\nID   Type    Date        Start     End       Project                Task\n@1   NonPom  2019-03-21  13:43:32  14:33:32  python                 Find the holy grail\n```\n\n\nHopefully this gave you a (not too confusing) introduction on the power of PomodoroWarrior. \n\nThere are many more command line arguments available than was illustrated in this example. Use the `--help` option for any command to learn its complete usage.  \n\n\n## Caveat ##\n\nTimeWarrior intervals does not have unique ids, and as such, PomodoroWarrior matches intervals through start and end times. \n\nIf intervals are modified with TimeWarrior directly, the link for that interval between the databases is lost.\n\nPomodoroWarrior tries to handle this situation gracefully, but it is recommend that you use PomodoroWarrior to modify any intervals, in order to keep the databases in sync.\n\n## Development ##\n\n### Build System ###\n\nStarting with version `0.0.3` we replace [setuptools](https://setuptools.readthedocs.io) with [poetry](https://poetry.eustace.io).\nPoetry is [PEP 518](https://www.python.org/dev/peps/pep-0518/) compliant, and uses `pyproject.toml` for configuration. In my opinion it provides for a simpler development experience.\n\nPoetry recommends that you install it isolated from the rest of your system. This can be done by running the install script:\n\n`curl -sSL https://raw.githubusercontent.com/sdispater/poetry/master/get-poetry.py | python`\n\n### Release ###\n\nThe script `./release.py` have been developed to automate cutting releases. It requires [git-extras](https://github.com/tj/git-extras) to be installed on your system.\n\nIf you want to manually create a release, you have to follow these rules:\n\n- Only release from `master`\n- Update `pomw/__version__.py` with the new version number\n- Update `pyproject.toml` with the new version number\n- Update `Changelog` with the commits since the previous release\n\n### Upload a release to PYPI ###\n\nOnce a tag is created, and pushed to Gitlab, the release is automatically uploaded to PYPI by the Gitlab CI publish stage.\n\n\n',
    'author': 'Tjaart van der Walt',
    'author_email': 'tjaart@tjaart.org',
    'url': 'https://tjaart.gitlab.io/pomw',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# -*- python-indent-offset: 4; -*-

from datetime import datetime

from .interval import Interval


class Pomodoro(Interval):
    def __init__(
        self,
        start_time=None,
        end_time=None,
        internal_interrupt=0,
        external_interrupt=0,
        void=False,
    ):
        super().__init__(start_time=start_time, end_time=end_time)
        self.internal_interrupt = internal_interrupt
        self.external_interrupt = external_interrupt
        self.void = void

    @staticmethod
    def __json_decode__(d):
        return Pomodoro(
            start_time=d.get("_Interval__start_time", None),
            end_time=d.get("_Interval__end_time", None),
            internal_interrupt=d.get("_Pomodoro__internal_interrupt", 0),
            external_interrupt=d.get("_Pomodoro__external_interrupt", 0),
            void=d.get("_Pomodoro__void", False),
        )

    @property
    def external_interrupt(self):
        return self.__external_interrupt

    @external_interrupt.setter
    def external_interrupt(self, external_interrupt):
        if isinstance(external_interrupt, int):
            self.__external_interrupt = external_interrupt
        else:
            raise TypeError(
                "'external_interrupt' should be of type %s" % (int.__name__)
            )

    @property
    def internal_interrupt(self):
        return self.__internal_interrupt

    @internal_interrupt.setter
    def internal_interrupt(self, internal_interrupt):
        if isinstance(internal_interrupt, int):
            self.__internal_interrupt = internal_interrupt
        else:
            raise TypeError(
                "'internal_interrupt' should be of type %s" % (int.__name__)
            )

    @property
    def void(self):
        return self.__void

    @void.setter
    def void(self, void):
        self.__void = void
